"""
Common definitions used everywhere for consistency.
https://plotly.com/python/reference/
"""
#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

# https://www.google.com/search?q=color+picker
GRID_COLOR = '#d1d1d1'
NODE_PRIMARY_COLOR = '#7ac4b7'
NODE_INTERNAL_COLOR = '#a832a8'
FRAME_COLOR = '#0f24db'
BEAM_MESH_COLOR = '#6d7aed'
FLOOR_COLOR = '#c4994f'
OFFSET_COLOR = '#a83256'
BISECTOR_COLOR = '#32a852'

if __name__ == "__main__":
    pass
